/*
  Copyright 2012-2013 by
  Centre de recherche INRIA Sophia-Antipolis Mediterranee, equipe APICS,
  Sophia Antipolis, France.

  Contributors S. Chevillard
  sylvain.chevillard@ens-lyon.org

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  It is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/


/* This program evaluates Brown's constant
                 B = Gamma(1/3)^6 * sqrt(10)/( 12 * pi^4 )
                   = 16*pi^2*sqrt(10)/( 81 * Gamma(2/3)^6 )
   See http://en.wikipedia.org/wiki/Particular_values_of_the_gamma_function
       http://www.iamned.com/math/

   Assume that s is a mpfr_t already initialized, and prec is a positive integer.
   After the call to mpfr_Browns_const(s, prec), the internal precision of s
   may have changed and s contains a value such that |B-s| \le 2^(1-prec)*B
*/

#include <mpfr.h>
#include "Brown.h"

#define MPFR_ACC_OR_MUL_UI(v)                            \
  do                                                    \
    {                                                   \
      if (v <= ULONG_MAX / acc)                         \
        acc *= v;                                       \
      else                                              \
        {                                               \
          mpfr_mul_ui (y, y, acc, mode); acc = v;       \
        }                                               \
    }                                                   \
  while (0)

#define MPFR_ACC_OR_DIV_UI(v)                            \
  do                                                    \
    {                                                   \
      if (v <= ULONG_MAX / acc)                         \
        acc *= v;                                       \
      else                                              \
        {                                               \
          mpfr_div_ui (y, y, acc, mode); acc = v;       \
        }                                               \
    }                                                   \
  while (0)

#define MPFR_ACC_OR_DIV_SI(v)                           \
  do                                                    \
    {                                                   \
      if (v <= LONG_MAX / acc)                          \
        acc *= v;                                       \
      else                                              \
        {                                               \
          mpfr_div_si (y, y, acc, mode); acc = v;       \
        }                                               \
    }                                                   \
  while (0)

void
mpfr_mul_ui2 (mpfr_ptr y, mpfr_srcptr x,
              unsigned long int v1, unsigned long int v2, mpfr_rnd_t mode)
{
  unsigned long int acc = v1;
  mpfr_set (y, x, mode);
  MPFR_ACC_OR_MUL_UI (v2);
  mpfr_mul_ui (y, y, acc, mode);
}


void
mpfr_mul_ui5 (mpfr_ptr y, mpfr_srcptr x,
              unsigned long int v1, unsigned long int v2,
              unsigned long int v3, unsigned long int v4,
              unsigned long int v5, mpfr_rnd_t mode)
{
  unsigned long int acc = v1;
  mpfr_set (y, x, mode);
  MPFR_ACC_OR_MUL_UI (v2);
  MPFR_ACC_OR_MUL_UI (v3);
  MPFR_ACC_OR_MUL_UI (v4);
  MPFR_ACC_OR_MUL_UI (v5);
  mpfr_mul_ui (y, y, acc, mode);
}

void
mpfr_div_ui2 (mpfr_ptr y, mpfr_srcptr x,
              unsigned long int v1, unsigned long int v2, mpfr_rnd_t mode)
{
  unsigned long int acc = v1;
  mpfr_set (y, x, mode);
  MPFR_ACC_OR_DIV_UI (v2);
  mpfr_div_ui (y, y, acc, mode);
}

void
mpfr_div_ui8 (mpfr_ptr y, mpfr_srcptr x,
              unsigned long int v1, unsigned long int v2,
              unsigned long int v3, unsigned long int v4,
              unsigned long int v5, unsigned long int v6,
              unsigned long int v7, unsigned long int v8, mpfr_rnd_t mode)
{
  unsigned long int acc = v1;
  mpfr_set (y, x, mode);
  MPFR_ACC_OR_DIV_UI (v2);
  MPFR_ACC_OR_DIV_UI (v3);
  MPFR_ACC_OR_DIV_UI (v4);
  MPFR_ACC_OR_DIV_UI (v5);
  MPFR_ACC_OR_DIV_UI (v6);
  MPFR_ACC_OR_DIV_UI (v7);
  MPFR_ACC_OR_DIV_UI (v8);
  mpfr_div_ui (y, y, acc, mode);
}

void
mpfr_div_si3 (mpfr_ptr y, mpfr_srcptr x,
              long int v1, long int v2,
              long int v3, mpfr_rnd_t mode)
{
  long int acc = v1;
  mpfr_set (y, x, mode);
  MPFR_ACC_OR_DIV_SI (v2);
  MPFR_ACC_OR_DIV_SI (v3);
  mpfr_div_si (y, y, acc, mode);
}



/* Gives an approximation of B = Gamma(1/3)^6 * sqrt(10) / (12pi^4)     */
/* using C. H. Brown's formula.                                         */
/* The computed value s satisfies |s-B| <= 2^{1-prec}*B                 */
/* As usual, the variable s is supposed to be initialized.              */
void
mpfr_Browns_const (mpfr_ptr s, mpfr_prec_t prec)
{
  mpfr_t uk;
  unsigned long int k;

  mpfr_prec_t working_prec = prec + 10 + MPFR_INT_CEIL_LOG2 (2 + (unsigned)prec / 10);

  mpfr_init2 (uk, working_prec);
  mpfr_set_prec (s, working_prec);

  mpfr_set_ui (uk, 1, GMP_RNDN);
  mpfr_set (s, uk, GMP_RNDN);
  k = 1;

  /* Invariants: uk ~ u(k-1) and s ~ sum(i=0..k-1, u(i)) */
  for (;;)
    {
      mpfr_mul_ui5 (uk, uk, 6*k-5, 6*k-4, 6*k-3, 6*k-2, 6*k-1, GMP_RNDN);
      mpfr_div_ui8 (uk, uk, k, k, 3*k-2, 3*k-1, 3*k, 80, 160, 160, GMP_RNDN);
      MPFR_CHANGE_SIGN (uk);

      mpfr_add (s, s, uk, GMP_RNDN);
      k++;
      if (MPFR_GET_EXP (uk) + (signed)(prec) <= MPFR_GET_EXP (s) + 7)
        break;
    }

  mpfr_clear (uk);
  return;
}
